import json
import pkgutil
import importlib
import os
import sys
import logging
from pathlib import Path

from . import expand_action
from .statement import Statement
from .finding import Finding
from .misc import make_list, ACCESS_DECISION


class Policy:
    _findings = []
    policy_json = None
    version = None
    statements = []
    policy = None

    def __init__(self, policy_json, filepath=None):
        self._findings = []
        self.statements = []
        self.policy_json = policy_json
        self.filepath = filepath

    def add_finding(self, finding, detail="", location={}):
        if "filepath" not in location:
            location["filepath"] = self.filepath
        self._findings.append(Finding(finding, detail, location))

    @property
    def findings(self):
        all_findings = []
        all_findings.extend(self._findings)

        for stmt in self.statements:
            for finding in stmt.findings:
                if "filepath" not in finding.location:
                    finding.location["filepath"] = self.filepath

                all_findings.append(finding)

        return all_findings

    @property
    def finding_ids(self):
        finding_ids = set()
        for finding in self.findings:
            finding_ids.add(finding.issue)
        return finding_ids

    @property
    def is_valid(self):
        for stmt in self.statements:
            if not stmt.is_valid:
                return False
        return True

    def get_references(self, privilege_prefix, privilege_name):
        """
        Identify all statements that reference this privilege,
        then return a dictionary where the keys are the resources referenced by the statements,
        and the values are a list of the statements
        """
        references = {}
        for stmt in self.statements:
            stmt_relevant_resources = stmt.get_resources_for_privilege(
                privilege_prefix, privilege_name
            )
            for resource in stmt_relevant_resources:
                references[resource] = references.get(resource, [])
                references[resource].append(stmt)
        return references

    def get_allowed_actions(self):
        actions_referenced = set()
        for stmt in self.statements:
            actions = make_list(stmt.stmt["Action"])
            for action in actions:
                expanded_actions = expand_action(action)
                for expanded_action in expanded_actions:
                    actions_referenced.add(
                        "{}:{}".format(
                            expanded_action["service"], expanded_action["action"]
                        )
                    )

        # actions_referenced is now a set like: {'lambda:UpdateFunctionCode', 'glue:UpdateDevEndpoint'}
        # We need to identify which of these are actually allowed though, as some of those could just be a deny
        # Worst case scenario though, we have a list of every action if someone included Action '*'

        allowed_actions = []
        for action in actions_referenced:
            parts = action.split(":")
            allowed_resources = self.get_allowed_resources(parts[0], parts[1])
            if len(allowed_resources) > 0:
                action = action.lower()
                allowed_actions.append(action)
        return allowed_actions

    def get_allowed_resources(self, privilege_prefix, privilege_name):
        """
        Given a privilege like s3:GetObject, identify all the resources (if any),
        this is allowed to be used with.

        Examples, assuming given "s3" "GetObject":
        - With a policy with s3:* on "*", this would return "*"
        - With a policy with s3:* on ["arn:aws:s3:::examplebucket", "arn:aws:s3:::examplebucket/*"],
          this would only return "arn:aws:s3:::examplebucket/*" because that is the only object resource.
        """

        def __is_allowed(stmts):
            """
            Given statements that are all relevant to the same resource and privilege,
            (meaning each statement must have an explicit allow or deny on the privilege) 
            determine if it is allowed, which means no Deny effects.
            """
            has_allow = False
            for stmt in stmts:
                if stmt.effect_allow:
                    has_allow = True
                else:
                    # If there is a Condition in the Deny, we don't count this as Deny'ing the action
                    # entirely so skip it
                    if "Condition" in stmt.stmt:
                        continue
                    return False
            return has_allow

        allowed_resources = []
        all_references = self.get_references(privilege_prefix, privilege_name)
        for resource in all_references:
            resource_is_allowed = __is_allowed(all_references[resource])

            # To avoid situations where we have an allow on a specific resource, but a deny
            # on *, I'm making a special case here
            # I should do regex intersections across each resource, but this will avoid
            # common situations for now
            if resource == "*" and not resource_is_allowed:
                # Only apply this case when the deny statement has no condition
                for stmt in all_references[resource]:
                    if not stmt.effect_allow and "Condition" not in stmt.stmt:
                        return []

            if resource_is_allowed:
                allowed_resources.append(resource)

        return allowed_resources

    def check_for_bad_patterns(self):
        """
        Look for privileges across multiple statements that result in problems such as privilege escalation.
        """

        def check_bucket_privesc(refs, bucket_privilege, object_privilege):
            # If the bucket privilege exists for a bucket, but not the object privilege for objects
            # in that bucket then the bucket privilege can be abused to get that object privilege
            for resource in refs[bucket_privilege]:
                if not (
                    resource in refs[object_privilege]
                    or resource + "/*" in refs[object_privilege]
                ):
                    self.add_finding(
                        "RESOURCE_POLICY_PRIVILEGE_ESCALATION",
                        detail="Possible resource policy privilege escalation on {} due to s3:{} not being allowed, but does allow s3:{}".format(
                            resource, object_privilege, bucket_privilege
                        ),
                    )

        # Get the resource references we'll be using
        refs = {}
        for priv in [
            "PutBucketPolicy",
            "PutBucketAcl",
            "PutLifecycleConfiguration",
            "PutObject",
            "GetObject",
            "DeleteObject",
        ]:
            refs[priv] = self.get_allowed_resources("s3", priv)

        # Check each bad combination.  If the bucket level privilege is allowed,
        # but not the object level privilege, then we likely have a privilege escalation issue.
        check_bucket_privesc(refs, "PutBucketPolicy", "PutObject")
        check_bucket_privesc(refs, "PutBucketPolicy", "GetObject")
        check_bucket_privesc(refs, "PutBucketPolicy", "DeleteObject")

        check_bucket_privesc(refs, "PutBucketAcl", "PutObject")
        check_bucket_privesc(refs, "PutBucketAcl", "GetObject")
        check_bucket_privesc(refs, "PutBucketAcl", "DeleteObject")

        check_bucket_privesc(refs, "PutLifecycleConfiguration", "DeleteObject")

    def analyze(
        self,
        ignore_private_auditors=False,
        private_auditors_custom_path=None,
        include_community_auditors=False,
    ):
        """
        Returns False if this policy is so broken that it couldn't be analyzed further.
        On True, it may still have findings.

        In either case, it will create Findings if there are any.
        """

        # Check no unknown elements exist
        for element in self.policy_json:
            if element not in ["Version", "Statement", "Id"]:
                self.add_finding(
                    "MALFORMED",
                    detail="Policy contains an unknown element",
                    location={"string": element},
                )
                return False

        # Check Version
        if "Version" not in self.policy_json:
            self.add_finding("NO_VERSION")
        else:
            self.version = self.policy_json["Version"]

            if self.version not in ["2012-10-17", "2008-10-17"]:
                self.add_finding("INVALID_VERSION", location={"string": self.version})
            elif self.version != "2012-10-17":
                # TODO I should have a check so that if an older version is being used,
                # and a variable is detected, it should be marked as higher severity.
                self.add_finding("OLD_VERSION", location={"string": self.version})

        # Check Statements
        if "Statement" not in self.policy_json:
            self.add_finding(
                "MALFORMED", detail="Policy does not contain a Statement element"
            )
            return False
        stmts_json = make_list(self.policy_json["Statement"])
        for stmt_json in stmts_json:
            stmt = Statement(stmt_json)
            self.statements.append(stmt)

        if not self.is_valid:
            # Do not continue. Further checks will not work with invalid statements.
            return False

        # Look for bad patterns
        self.check_for_bad_patterns()

        if not ignore_private_auditors:
            # Import any private auditing modules
            private_auditors_directory = "private_auditors"
            private_auditors_directory_path = (
                Path(os.path.abspath(__file__)).parent / private_auditors_directory
            )

            if private_auditors_custom_path is not None:
                private_auditors_directory_path = private_auditors_custom_path
                # Ensure we can import from this directory
                sys.path.append(".")

            private_auditors = {}
            for importer, name, _ in pkgutil.iter_modules(
                [private_auditors_directory_path]
            ):
                full_package_name = "parliament.%s.%s" % (
                    private_auditors_directory,
                    name,
                )

                if private_auditors_custom_path is not None:
                    path_with_dots = private_auditors_directory_path.replace(
                        "/", "."
                    ).replace("\\", ".")
                    full_package_name = path_with_dots + "." + name

                module = importlib.import_module(full_package_name)
                private_auditors[name] = module

            if len(private_auditors) == 0 and private_auditors_custom_path is not None:
                raise Exception(
                    "No private auditors found at {}".format(
                        private_auditors_custom_path
                    )
                )

            # Run them
            for m in private_auditors:
                logging.info(f"*** Checking with private auditor: {m}")
                private_auditors[m].audit(self)

        if include_community_auditors is True:
            # Import any private auditing modules
            community_auditors_directory = "community_auditors"
            community_auditors_directory_path = (
                Path(os.path.abspath(__file__)).parent / community_auditors_directory
            )

            community_auditors = {}
            for importer, name, _ in pkgutil.iter_modules(
                [community_auditors_directory_path]
            ):
                full_package_name = "parliament.%s.%s" % (
                    community_auditors_directory,
                    name,
                )

                path_with_dots = full_package_name.replace("/", ".")
                full_package_name = path_with_dots

                module = importlib.import_module(full_package_name)
                community_auditors[name] = module

            # Run them
            for m in community_auditors:
                logging.info(f"*** Checking with community auditor: {m}")
                community_auditors[m].audit(self)

        return True



from parliament import is_arn_match, expand_action

# https://gist.github.com/kmcquade/33860a617e651104d243c324ddf7992a
CREDENTIALS_EXPOSURE_ACTIONS = [
    "codepipeline:pollforjobs",
    "cognito-idp:associatesoftwaretoken",
    "cognito-identity:getopenidtoken",
    "cognito-identity:getopenidtokenfordeveloperidentity",
    "cognito-identity:getcredentialsforidentity",
    "connect:getfederationtoken",
    "connect:getfederationtokens",
    "ecr:getauthorizationtoken",
    "gamelift:requestuploadcredentials",
    "iam:createaccesskey",
    "iam:createloginprofile",
    "iam:createservicespecificcredential",
    "iam:resetservicespecificcredential",
    "iam:updateaccesskey",
    "iot:assumerolewithcertificate",
    "lightsail:getinstanceaccessdetails",
    "lightsail:getrelationaldatabasemasteruserpassword",
    "rds-db:connect",
    "redshift:getclustercredentials",
    "sso:getrolecredentials",
    "mediapackage:rotateingestendpointcredentials",
    "sts:assumerole",
    "sts:assumerolewithsaml",
    "sts:assumerolewithwebidentity",
    "sts:getfederationtoken",
    "sts:getsessiontoken",
]


def audit(policy):
    actions = policy.get_allowed_actions()

    credentials_exposure_actions_in_policy = []
    for action in actions:
        if action in CREDENTIALS_EXPOSURE_ACTIONS:
            credentials_exposure_actions_in_policy.append(action)
    if len(credentials_exposure_actions_in_policy) > 0:
        policy.add_finding(
            "CREDENTIALS_EXPOSURE",
            location={"actions": credentials_exposure_actions_in_policy},
        )



#!/usr/bin/env python3
import argparse
import json
import logging
import re
import sys
from os import walk
from os.path import abspath
from os.path import join
from pathlib import Path

from parliament import (
    analyze_policy_string,
    enhance_finding,
    override_config,
    __version__,
)
from parliament.misc import make_list

logger = logging.getLogger(__name__)


def is_finding_filtered(finding, minimum_severity="LOW"):
    # Return True if the finding should be filtered (ie. return False if it should be displayed)
    minimum_severity = minimum_severity.upper()
    severity_choices = ["MUTE", "INFO", "LOW", "MEDIUM", "HIGH", "CRITICAL"]
    if severity_choices.index(finding.severity) < severity_choices.index(
        minimum_severity
    ):
        return True

    if finding.ignore_locations:
        # The ignore_locations element looks like this:
        #
        # ignore_locations:
        # - filepath: "test.json"
        #   action: "s3:GetObject"
        #   resource:
        #   - "a"
        #   - "b"
        # - action: "s3:GetObject"
        #   resource:
        #    - "c.*"
        #
        # Assuming the finding has these types of values in the `location` element,
        # this will ignore any finding that matches the filepath to "test.json"
        # AND action to "s3:GetObject"
        # AND the resource to "a" OR "b"
        # It will also ignore a resource that matches "c.*".

        for ignore_location in finding.ignore_locations:
            all_match = True
            for location_type, locations_to_ignore in ignore_location.items():
                has_array_match = False
                for location_to_ignore in make_list(locations_to_ignore):
                    if re.fullmatch(
                        location_to_ignore.lower(),
                        str(finding.location.get(location_type, "")).lower(),
                    ):
                        has_array_match = True
                if not has_array_match:
                    all_match = False
            if all_match:
                return True
    return False


def print_finding(finding, minimal_output=False, json_output=False):
    if minimal_output:
        print("{}".format(finding.issue))
    elif json_output:
        print(
            json.dumps(
                {
                    "issue": finding.issue,
                    "title": finding.title,
                    "severity": finding.severity,
                    "description": finding.description,
                    "detail": finding.detail,
                    "location": finding.location,
                }
            )
        )
    else:
        print(
            "{} - {} - {} - {} - {}".format(
                finding.severity,
                finding.title,
                finding.description,
                finding.detail,
                finding.location,
            )
        )


def find_files(directory, exclude_pattern=None, policy_extension=""):
    exclude = None
    if exclude_pattern:
        exclude = re.compile(exclude_pattern)

    discovered_files = []
    for root, _, files in walk(directory):
        for name in files:
            if name.endswith(policy_extension):
                file = join(root, name)
                if exclude:
                    result = exclude.match(file)
                    if result:
                        logger.info(
                            'Found file %s matches exclude pattern "%s"',
                            file,
                            exclude_pattern,
                        )
                        continue
                logger.info("Found file %s", file)
                discovered_files.append(file)

    return discovered_files


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--aws-managed-policies",
        help="This is used with the policies directory of https://github.com/SummitRoute/aws_managed_policies",
        type=str,
    )
    parser.add_argument(
        "--auth-details-file",
        help='Provide the path to a file returned by "aws iam get-account-authorization-details"',
        type=str,
    )
    parser.add_argument(
        "--string",
        help='Provide a string such as \'{"Version": "2012-10-17","Statement": {"Effect": "Allow","Action": ["s3:GetObject", "s3:PutBucketPolicy"],"Resource": ["arn:aws:s3:::bucket1", "arn:aws:s3:::bucket2/*"]}}\'',
        type=str,
    )
    parser.add_argument("--file", help="Provide a policy in a file", type=str)
    parser.add_argument(
        "--directory", help="Provide a path to directory with policy files", type=str
    )
    parser.add_argument(
        "--include_policy_extension",
        help="Policy file extension to scan for in directory mode (ex. \".json\")",
        default="json",
        type=str,
    )
    parser.add_argument(
        "--exclude_pattern", help="File name regex pattern to exclude (ex. \".*venv.*\")", type=str
    )
    parser.add_argument(
        "--minimal", help="Minimal output", default=False, action="store_true"
    )
    parser.add_argument(
        "--json", help="json output", default=False, action="store_true"
    )
    parser.add_argument(
        "--minimum_severity",
        help="Minimum severity to display. Options: CRITICAL, HIGH, MEDIUM, LOW, INFO",
        default="LOW",
        choices=["CRITICAL", "HIGH", "MEDIUM", "LOW", "INFO"],
    )
    parser.add_argument(
        "--private_auditors",
        help="Directory of the private auditors. Defaults to looking in private_auditors in the same directory as the iam_definition.json file.",
        default=None,
    )
    parser.add_argument(
        "--config", help="Custom config file for over-riding values", type=str
    )
    parser.add_argument(
        "--include-community-auditors",
        help="Use this flag to enable community-provided auditors",
        default=None,
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="Increase output verbosity",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--version",
        action="version",
        version="%(prog)s {version}".format(version=__version__),
    )
    args = parser.parse_args()

    log_level = logging.ERROR
    log_format = "%(message)s"

    # We want to silence dependencies
    logging.getLogger("botocore").setLevel(logging.CRITICAL)
    logging.getLogger("boto3").setLevel(logging.CRITICAL)
    logging.getLogger("urllib3").setLevel(logging.CRITICAL)

    if args.verbose:
        log_level = logging.INFO
        log_format = "%(message)s"

    logging.basicConfig(level=log_level, stream=sys.stderr, format=log_format)

    if args.private_auditors is not None and "." in args.private_auditors:
        raise Exception("The path to the private_auditors must not have periods")

    if args.minimal and args.json:
        raise Exception("You cannot choose both minimal and json output")

    # Change the exit status if there are errors
    exit_status = 0
    findings = []

    if args.aws_managed_policies:
        file_paths = find_files(directory=args.aws_managed_policies)
        for file_path in file_paths:
            with open(file_path) as f:
                contents = f.read()
                policy_file_json = json.loads(contents)
                policy_string = json.dumps(
                    policy_file_json["PolicyVersion"]["Document"]
                )
                policy = analyze_policy_string(
                    policy_string,
                    file_path,
                    private_auditors_custom_path=args.private_auditors,
                    include_community_auditors=args.include_community_auditors,
                )
                findings.extend(policy.findings)

    elif args.auth_details_file:
        with open(args.auth_details_file) as f:
            contents = f.read()
            auth_details_json = json.loads(contents)
            for policy in auth_details_json["Policies"]:
                # Ignore AWS defined policies
                if "arn:aws:iam::aws:" not in policy["Arn"]:
                    continue

                # Ignore AWS Service-linked roles
                if (
                    policy["Path"] == "/service-role/"
                    or policy["Path"] == "/aws-service-role/"
                    or policy["PolicyName"].startswith("AWSServiceRoleFor")
                    or policy["PolicyName"].endswith("ServiceRolePolicy")
                    or policy["PolicyName"].endswith("ServiceLinkedRolePolicy")
                ):
                    continue

                for version in policy["PolicyVersionList"]:
                    if not version["IsDefaultVersion"]:
                        continue
                    policy = analyze_policy_string(
                        json.dumps(version["Document"]), policy["Arn"],
                    )
                    findings.extend(policy.findings)

            # Review the inline policies on Users, Roles, and Groups
            for user in auth_details_json["UserDetailList"]:
                for policy in user.get("UserPolicyList", []):
                    policy = analyze_policy_string(
                        json.dumps(policy["PolicyDocument"]),
                        user["Arn"],
                        private_auditors_custom_path=args.private_auditors,
                        include_community_auditors=args.include_community_auditors,
                    )
                    findings.extend(policy.findings)
            for role in auth_details_json["RoleDetailList"]:
                for policy in role.get("RolePolicyList", []):
                    policy = analyze_policy_string(
                        json.dumps(policy["PolicyDocument"]),
                        role["Arn"],
                        private_auditors_custom_path=args.private_auditors,
                        include_community_auditors=args.include_community_auditors,
                    )
                    findings.extend(policy.findings)
            for group in auth_details_json["GroupDetailList"]:
                for policy in group.get("GroupPolicyList", []):
                    policy = analyze_policy_string(
                        json.dumps(policy["PolicyDocument"]),
                        group["Arn"],
                        private_auditors_custom_path=args.private_auditors,
                        include_community_auditors=args.include_community_auditors,
                    )
                    findings.extend(policy.findings)
    elif args.string:
        policy = analyze_policy_string(
            args.string,
            private_auditors_custom_path=args.private_auditors,
            include_community_auditors=args.include_community_auditors,
        )
        findings.extend(policy.findings)
    elif args.file:
        with open(args.file) as f:
            contents = f.read()
            policy = analyze_policy_string(
                contents,
                args.file,
                private_auditors_custom_path=args.private_auditors,
                include_community_auditors=args.include_community_auditors,
            )
            findings.extend(policy.findings)
    elif args.directory:
        file_paths = find_files(
            directory=args.directory,
            exclude_pattern=args.exclude_pattern,
            policy_extension=args.include_policy_extension,
        )
        for file_path in file_paths:
            with open(file_path) as f:
                contents = f.read()
                policy = analyze_policy_string(
                    contents,
                    file_path,
                    private_auditors_custom_path=args.private_auditors,
                    include_community_auditors=args.include_community_auditors,
                )
                findings.extend(policy.findings)
    else:
        parser.print_help()
        exit(-1)

    filtered_findings = []
    override_config(args.config)
    if args.include_community_auditors:
        community_auditors_directory = "community_auditors"
        community_auditors_override_file = (
            Path(abspath(__file__)).parent
            / community_auditors_directory
            / "config_override.yaml"
        )
        override_config(community_auditors_override_file)
    for finding in findings:
        finding = enhance_finding(finding)
        if not is_finding_filtered(finding, args.minimum_severity):
            filtered_findings.append(finding)

    if len(filtered_findings) == 0:
        # Return with exit code 0 if no findings
        return

    for finding in filtered_findings:
        print_finding(finding, args.minimal, args.json)

    # There were findings, so return with a non-zero exit code
    exit(1)


if __name__ == "__main__":
    main()



def make_list(v):
    """
    If the object is not a list already, it converts it to one
    Examples:
    [1, 2, 3] -> [1, 2, 3]
    [1] -> [1]
    1 -> [1]
    """
    if not isinstance(v, list):
        return [v]
    return v


class ACCESS_DECISION:
    IMPLICIT_DENY = 0
    EXPLICIT_DENY = 1
    EXPLICIT_ALLOW = 2
